# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
The module checks that SMI events configuration is locked down
- Global SMI Enable/SMI Lock
- TCO SMI Enable/TCO Lock

References:
    - `Setup for Failure: Defeating SecureBoot <http://syscan.org/index.php/download/get/6e597f6067493dd581eed737146f3afb/SyScan2014_CoreyKallenberg_SetupforFailureDefeatingSecureBoot.zip>`_ by Corey Kallenberg, Xeno Kovah, John Butterworth, Sam Cornwell
    - `Summary of Attacks Against BIOS and Secure Boot <https://www.defcon.org/images/defcon-22/dc-22-presentations/Bulygin-Bazhaniul-Furtak-Loucaides/DEFCON-22-Bulygin-Bazhaniul-Furtak-Loucaides-Summary-of-attacks-against-BIOS-UPDATED.pdf>`_

Usage:
    ``chipsec_main -m common.bios_smi``

Examples:
    >>> chipsec_main.py -m common.bios_smi

Registers used:
    - SmmBiosWriteProtection (Control)
    - TCOSMILock (Control)
    - SMILock (Control)
    - BiosWriteEnable (Control)

"""

from chipsec.module_common import BaseModule, BIOS, SMM
from chipsec.library.returncode import ModuleResult
from typing import List

TAGS = [BIOS, SMM]
METADATA_TAGS = ['OPENSOURCE', 'IA', 'COMMON', 'BIOS_SMI']


class bios_smi(BaseModule):

    def __init__(self):
        BaseModule.__init__(self)

    def is_supported(self) -> bool:
        if not self.cs.control.is_defined('SmmBiosWriteProtection') or \
           not self.cs.control.is_defined('TCOSMILock') or \
           not self.cs.control.is_defined('SMILock') or \
           not self.cs.control.is_defined('BiosWriteEnable'):
            self.logger.log_important('Required controls not defined for platform.  Skipping module.')
            return False
        return True

    def check_SMI_locks(self) -> int:

        smm_bwp = self.cs.control.get('SmmBiosWriteProtection')
        if 0 == smm_bwp:
            self.logger.log_bad('SMM BIOS region write protection has not been enabled (SMM_BWP is not used)\n')
        else:
            self.logger.log_good('SMM BIOS region write protection is enabled (SMM_BWP is used)\n')

        ok = True
        warn = False

        if self.cs.control.is_defined('TCOSMIEnable') and self.cs.control.is_defined('GlobalSMIEnable'):
            self.logger.log('[*] Checking SMI enables..')
            tco_en = self.cs.control.get('TCOSMIEnable')
            gbl_smi_en = self.cs.control.get('GlobalSMIEnable')
            self.logger.log(f'    Global SMI enable: {gbl_smi_en:d}')
            self.logger.log(f'    TCO SMI enable   : {tco_en:d}')

            if gbl_smi_en != 1:
                ok = False
                self.logger.log_bad('Global SMI is not enabled')
            elif (tco_en != 1) and (smm_bwp != 1):
                warn = True
                self.logger.log_warning('TCO SMI is not enabled. BIOS may not be using it')
            elif (tco_en != 1) and (smm_bwp == 1):
                ok = False
                self.logger.log_bad('TCO SMI should be enabled if using SMM BIOS region protection')
            else:
                self.logger.log_good('All required SMI events are enabled')
            self.logger.log('')
            self.logger.log('[*] Checking SMI configuration locks..')

        tco_lock = self.cs.control.get('TCOSMILock')
        if tco_lock != 1:
            ok = False
            self.logger.log_bad('TCO SMI event configuration is not locked. TCO SMI events can be disabled')
        else:
            self.logger.log_good('TCO SMI configuration is locked (TCO SMI Lock)')

        smi_lock = self.cs.control.get('SMILock')
        if smi_lock != 1:
            ok = False
            self.logger.log_bad('SMI events global configuration is not locked. SMI events can be disabled')
        else:
            self.logger.log_good('SMI events global configuration is locked (SMI Lock)')
        self.logger.log('')

        if ok and not warn:
            res = ModuleResult.PASSED
            self.logger.log_passed('All required SMI sources seem to be enabled and locked')
        elif ok and warn:
            res = ModuleResult.WARNING
            self.result.setStatusBit(self.result.status.VERIFY)
            self.logger.log_warning('One or more warnings detected when checking SMI enable state')
        else:
            res = ModuleResult.FAILED
            self.result.setStatusBit(self.result.status.LOCKS)
            self.logger.log_failed('Not all required SMI sources are enabled and locked')

        return self.result.getReturnCode(res)

    def run(self, module_argv: List[str]) -> int:
        self.logger.start_test('SMI Events Configuration')
        self.res = self.check_SMI_locks()
        return self.res



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2023, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

from typing import Optional

BIT0 = 0x0001
BIT1 = 0x0002
BIT2 = 0x0004
BIT3 = 0x0008
BIT4 = 0x0010
BIT5 = 0x0020
BIT6 = 0x0040
BIT7 = 0x0080
BIT8 = 0x0100
BIT9 = 0x0200
BIT10 = 0x0400
BIT11 = 0x0800
BIT12 = 0x1000
BIT13 = 0x2000
BIT14 = 0x4000
BIT15 = 0x8000
BIT16 = 0x00010000
BIT17 = 0x00020000
BIT18 = 0x00040000
BIT19 = 0x00080000
BIT20 = 0x00100000
BIT21 = 0x00200000
BIT22 = 0x00400000
BIT23 = 0x00800000
BIT24 = 0x01000000
BIT25 = 0x02000000
BIT26 = 0x04000000
BIT27 = 0x08000000
BIT28 = 0x10000000
BIT29 = 0x20000000
BIT30 = 0x40000000
BIT31 = 0x80000000
BIT32 = 0x100000000
BIT33 = 0x200000000
BIT34 = 0x400000000
BIT35 = 0x800000000
BIT36 = 0x1000000000
BIT37 = 0x2000000000
BIT38 = 0x4000000000
BIT39 = 0x8000000000
BIT40 = 0x10000000000
BIT41 = 0x20000000000
BIT42 = 0x40000000000
BIT43 = 0x80000000000
BIT44 = 0x100000000000
BIT45 = 0x200000000000
BIT46 = 0x400000000000
BIT47 = 0x800000000000
BIT48 = 0x1000000000000
BIT49 = 0x2000000000000
BIT50 = 0x4000000000000
BIT51 = 0x8000000000000
BIT52 = 0x10000000000000
BIT53 = 0x20000000000000
BIT54 = 0x40000000000000
BIT55 = 0x80000000000000
BIT56 = 0x100000000000000
BIT57 = 0x200000000000000
BIT58 = 0x400000000000000
BIT59 = 0x800000000000000
BIT60 = 0x1000000000000000
BIT61 = 0x2000000000000000
BIT62 = 0x4000000000000000
BIT63 = 0x8000000000000000


def make_mask(size: int, mask_start: int = 0) -> int:
    mask = (1 << size) - 1
    mask <<= mask_start
    return mask


def bit(bit_num: int) -> int:
    return int(1 << bit_num)


def is_set(val: int, bit_mask: int) -> bool:
    return bool(val & bit_mask != 0)


def scan_single_bit_mask(bit_mask: int) -> Optional[int]:
    for bit_num in range(0, 7):
        if bit_mask >> bit_num == 1:
            return bit_num
    return None


def is_all_ones(value: int, size: int, width: int = 8) -> bool:
    mask = (1 << (size * width)) - 1
    return (mask == (mask & value))


def ones_complement(value: int, number_of_bits: int = 64) -> int:
    return ((1 << number_of_bits) - 1) ^ value


def get_bits(value: int, start: int, nbits: int) -> int:
    ret = value >> start
    ret &= (1 << nbits) - 1
    return ret


def set_bits(bit: int, size: int, initial_value: int, value: int) -> int:
    field_mask = make_mask(size)
    new_value = initial_value & ~(field_mask << bit)
    new_value |= ((value & field_mask) << bit)
    return new_value



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2023, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

# Usage: python strip_record_json_of_pcienumeration.py recording.json
#     Where recording.json is generated from using "--helper recordhelper" on 
#     chipsec_main.py or chipsec_util.py

import sys
import os
from json import dumps, loads
class Split_Enumeration:
    def __init__(self, filename:str = "") -> None:
        self.data = {}
        self.enum_data = {}
        self.set_filename(filename)
        if self.filename:
            self.load_data()
        
    def set_filename(self, filename: str) -> None:
        self.filename = os.path.splitext(filename)[0]
        
    def load_data(self) -> None:
        if not self.filename:
            print("No file specified.")
            sys.exit(1)
        with open(self.filename+".json") as f:
            rawdata = f.read()
        self.data = loads(rawdata)


    def split_enumeration(self) -> None:
        fs_as_base_10_str = "4294967295" #0xFFFFFFFF
        
        self.pop_add_del_if_empty("cpuid", "(1,0)")
        
        cmd = "read_pci_reg"
        for b in range(256):
            for d in range(32):
                for f in range(8):
                    targs = f"({b},{d},{f},0,4)"
                    popdata = self.pop_add_del_if_empty(cmd, targs)
                    if f == 0 and popdata == fs_as_base_10_str:
                        break
                    if popdata != fs_as_base_10_str:
                        targs = f"({b},{d},{f},8,1)"
                        self.pop_add_del_if_empty(cmd, targs)

        self.save_all_jsons()

    def pop_add_del_if_empty(self, cmd:str, targs:str) -> str:
        popdata = self.data[cmd][targs].pop()
        self.add_to_enum_data(cmd, targs, popdata)
        if len(self.data[cmd][targs]) == 0:
            del self.data[cmd][targs]
            if len(self.data[cmd]) == 0:
                del self.data[cmd]
        return popdata

    def add_to_enum_data(self, cmd:str, targs:str, popdata:str) -> None:
        if cmd not in self.enum_data:
            self.enum_data[cmd] = {}
        if targs not in self.enum_data[cmd]:
            self.enum_data[cmd][targs] = []
        self.enum_data[cmd][targs].insert(0, popdata)

    def save_all_jsons(self) -> None:
        self.save_json(self.data, "test")
        self.save_json(self.enum_data, "enum")

    def save_json(self, data:dict, suffix:str) -> None:
        js = dumps(data, sort_keys=False, indent=2, separators=(',', ': ')) 
        with open(f"{self.filename}_{suffix}.json", 'w') as f:
            f.write(js)

if __name__ == '__main__':
    if len(sys.argv) == 2:
        Split_Enumeration(sys.argv[1]).split_enumeration()
    else:
        print("Expecting one argument: filename")



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
Management and communication with Windows kernel mode driver which provides access to hardware resources

.. note::
    On Windows you need to install pywin32 Python extension corresponding to your Python version:
    https://github.com/mhammond/pywin32
"""

import errno
import os.path
import platform
import pywintypes
import struct
import sys
import winerror
import win32service
import win32api, win32process, win32security, win32serviceutil, win32file
from collections import namedtuple
from ctypes import windll, Structure, pythonapi, py_object,  Array, POINTER
from ctypes import addressof, sizeof, create_string_buffer, WinError
from ctypes import c_ulong, c_ushort, c_char_p, c_size_t, c_int, c_uint32, c_wchar_p, c_void_p, c_char
from typing import Dict, List, Optional, Tuple, AnyStr, TYPE_CHECKING
from win32file import FILE_SHARE_READ, FILE_SHARE_WRITE, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, FILE_FLAG_OVERLAPPED, INVALID_HANDLE_VALUE
from win32.lib import win32con
if TYPE_CHECKING:
    from pywintypes import PyHANDLE

from chipsec.library.exceptions import OsHelperError, HWAccessViolationError, UnimplementedAPIError
from chipsec.helper.basehelper import Helper
from chipsec.library.defines import stringtobytes, bytestostring
from chipsec.library.logger import logger
import chipsec.library.file
from chipsec.hal.uefi_common import EFI_GUID_STR


class PCI_BDF(Structure):
    _fields_ = [("BUS", c_ushort, 16),  # Bus
                ("DEV", c_ushort, 16),  # Device
                ("FUNC", c_ushort, 16),  # Function
                ("OFF", c_ushort, 16)]  # Offset


kernel32 = windll.kernel32


drv_hndl_error_msg = "Cannot open chipsec driver handle. Make sure chipsec driver is installed and started if you are using option -e (see README)"

DRIVER_FILE_PATHS = [os.path.join("C:\\", "Windows", "System32", "drivers"), os.path.join(chipsec.library.file.get_main_dir(), "chipsec", "helper", "windows", f'windows_{platform.machine().lower()}')]
DRIVER_FILE_NAME = "chipsec_hlpr.sys"
DEVICE_FILE = "\\\\.\\chipsec_hlpr"
SERVICE_NAME = "chipsec"
DISPLAY_NAME = "CHIPSEC Service"

CHIPSEC_INSTALL_PATH = os.path.join(sys.prefix, "Lib", "site-packages", "chipsec")

# Status Codes
STATUS_PRIVILEGED_INSTRUCTION = 0xC0000096

# Defines for Win32 API Calls
GENERIC_READ = 0x80000000
GENERIC_WRITE = 0x40000000
OPEN_EXISTING = 0x3

FILE_DEVICE_UNKNOWN = 0x00000022

METHOD_BUFFERED = 0
METHOD_IN_DIRECT = 1
METHOD_OUT_DIRECT = 2
METHOD_NEITHER = 3

FILE_ANY_ACCESS: int = 0
FILE_SPECIAL_ACCESS = (FILE_ANY_ACCESS)
FILE_READ_ACCESS = (0x0001)
FILE_WRITE_ACCESS = (0x0002)


def CTL_CODE(DeviceType: int, Function: int, Method: int, Access: int) -> int:
    return ((DeviceType) << 16) | ((Access) << 14) | ((Function) << 2) | (Method)


#
# chipsec driver IOCTL codes
#
CHIPSEC_CTL_ACCESS: int = (FILE_READ_ACCESS | FILE_WRITE_ACCESS)

CLOSE_DRIVER: int = CTL_CODE(FILE_DEVICE_UNKNOWN, 0x803, METHOD_BUFFERED, CHIPSEC_CTL_ACCESS)
READ_PCI_CFG_REGISTER: int = CTL_CODE(FILE_DEVICE_UNKNOWN, 0x807, METHOD_BUFFERED, CHIPSEC_CTL_ACCESS)
WRITE_PCI_CFG_REGISTER: int = CTL_CODE(FILE_DEVICE_UNKNOWN, 0x808, METHOD_BUFFERED, CHIPSEC_CTL_ACCESS)
IOCTL_READ_PHYSMEM: int = CTL_CODE(FILE_DEVICE_UNKNOWN, 0x809, METHOD_BUFFERED, CHIPSEC_CTL_ACCESS)
IOCTL_WRITE_PHYSMEM: int = CTL_CODE(FILE_DEVICE_UNKNOWN, 0x80a, METHOD_BUFFERED, CHIPSEC_CTL_ACCESS)
IOCTL_LOAD_UCODE_PATCH: int = CTL_CODE(FILE_DEVICE_UNKNOWN, 0x80b, METHOD_BUFFERED, CHIPSEC_CTL_ACCESS)
IOCTL_WRMSR: int = CTL_CODE(FILE_DEVICE_UNKNOWN, 0x80c, METHOD_BUFFERED, CHIPSEC_CTL_ACCESS)
IOCTL_RDMSR: int = CTL_CODE(FILE_DEVICE_UNKNOWN, 0x80d, METHOD_BUFFERED, CHIPSEC_CTL_ACCESS)
IOCTL_READ_IO_PORT: int = CTL_CODE(FILE_DEVICE_UNKNOWN, 0x80e, METHOD_BUFFERED, CHIPSEC_CTL_ACCESS)
IOCTL_WRITE_IO_PORT: int = CTL_CODE(FILE_DEVICE_UNKNOWN, 0x80f, METHOD_BUFFERED, CHIPSEC_CTL_ACCESS)
IOCTL_GET_CPU_DESCRIPTOR_TABLE: int = CTL_CODE(FILE_DEVICE_UNKNOWN, 0x810, METHOD_BUFFERED, CHIPSEC_CTL_ACCESS)
IOCTL_SWSMI: int = CTL_CODE(FILE_DEVICE_UNKNOWN, 0x811, METHOD_BUFFERED, CHIPSEC_CTL_ACCESS)
IOCTL_ALLOC_PHYSMEM: int = CTL_CODE(FILE_DEVICE_UNKNOWN, 0x812, METHOD_BUFFERED, CHIPSEC_CTL_ACCESS)
IOCTL_CPUID: int = CTL_CODE(FILE_DEVICE_UNKNOWN, 0x813, METHOD_BUFFERED, CHIPSEC_CTL_ACCESS)
IOCTL_HYPERCALL: int = CTL_CODE(FILE_DEVICE_UNKNOWN, 0x814, METHOD_BUFFERED, CHIPSEC_CTL_ACCESS)
IOCTL_GET_PHYSADDR: int = CTL_CODE(FILE_DEVICE_UNKNOWN, 0x815, METHOD_BUFFERED, CHIPSEC_CTL_ACCESS)
IOCTL_MAP_IO_SPACE: int = CTL_CODE(FILE_DEVICE_UNKNOWN, 0x816, METHOD_BUFFERED, CHIPSEC_CTL_ACCESS)
IOCTL_FREE_PHYSMEM: int = CTL_CODE(FILE_DEVICE_UNKNOWN, 0x817, METHOD_BUFFERED, CHIPSEC_CTL_ACCESS)
IOCTL_WRCR: int = CTL_CODE(FILE_DEVICE_UNKNOWN, 0x818, METHOD_BUFFERED, CHIPSEC_CTL_ACCESS)
IOCTL_RDCR: int = CTL_CODE(FILE_DEVICE_UNKNOWN, 0x819, METHOD_BUFFERED, CHIPSEC_CTL_ACCESS)
IOCTL_MSGBUS_SEND_MESSAGE: int = CTL_CODE(FILE_DEVICE_UNKNOWN, 0x81a, METHOD_BUFFERED, CHIPSEC_CTL_ACCESS)
IOCTL_WRITE_MMIO: int = CTL_CODE(FILE_DEVICE_UNKNOWN, 0x81b, METHOD_BUFFERED, CHIPSEC_CTL_ACCESS)
IOCTL_READ_MMIO: int = CTL_CODE(FILE_DEVICE_UNKNOWN, 0x81c, METHOD_BUFFERED, CHIPSEC_CTL_ACCESS)

#
# Format for IOCTL Structures
#
_pack = 'Q' if sys.maxsize > 2**32 else 'I'
_smi_msg_t_fmt = 7 * _pack

#
# NT Errors
#
# Defined in WinDDK\7600.16385.1\inc\api\ntstatus.h
#

#
# UEFI constants
#
# Default buffer size for EFI variables
# EFI_VAR_MAX_BUFFER_SIZE = 128*1024
EFI_VAR_MAX_BUFFER_SIZE = 1024 * 1024

attributes: Dict[str, int] = {
    "EFI_VARIABLE_NON_VOLATILE": 0x00000001,
    "EFI_VARIABLE_BOOTSERVICE_ACCESS": 0x00000002,
    "EFI_VARIABLE_RUNTIME_ACCESS": 0x00000004,
    "EFI_VARIABLE_HARDWARE_ERROR_RECORD": 0x00000008,
    "EFI_VARIABLE_AUTHENTICATED_WRITE_ACCESS": 0x00000010,
    "EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS": 0x00000020,
    "EFI_VARIABLE_APPEND_WRITE": 0x00000040
}

PyLong_AsByteArray = pythonapi._PyLong_AsByteArray
PyLong_AsByteArray.argtypes = [py_object,
                               c_char_p,
                               c_size_t,
                               c_int,
                               c_int]


def packl_ctypes(lnum: int, bitlength: int) -> bytes:
    length = (bitlength + 7) // 8
    a = create_string_buffer(length)
    PyLong_AsByteArray(lnum, a, len(a), 1, 1)  # 4th param is for endianness 0 - big, non 0 - little
    return a.raw


#
# Firmware Table Provider Signatures
#
FirmwareTableProviderSignature_ACPI = 0x41435049  # 'ACPI' - The ACPI firmware table provider
FirmwareTableProviderSignature_FIRM = 0x4649524D  # 'FIRM' - The raw firmware table provider
FirmwareTableProviderSignature_RSMB = 0x52534D42  # 'RSMB' - The raw SMBIOS firmware table provider

FirmwareTableID_RSDT = 0x54445352
FirmwareTableID_XSDT = 0x54445358


class EFI_HDR_WIN(namedtuple('EFI_HDR_WIN', 'Size DataOffset DataSize Attributes guid')):
    __slots__ = ()

    def __str__(self) -> str:
        return f"""
Header (Windows)
----------------
VendorGuid= {{{EFI_GUID_STR(self.guid)}}}
Size      = 0x{self.Size:08X}
DataOffset= 0x{self.DataOffset:08X}
DataSize  = 0x{self.DataSize:08X}
Attributes= 0x{self.Attributes:08X}
"""


def getEFIvariables_NtEnumerateSystemEnvironmentValuesEx2(nvram_buf: bytes) -> Dict[str, List[Tuple[int, bytes, int, bytes, str, int]]]:
    start = 0
    buffer = nvram_buf
    bsize = len(buffer)
    header_fmt = "<IIII16s"
    header_size = struct.calcsize(header_fmt)
    variables = dict()
    off = 0
    while (off + header_size) < bsize:
        efi_var_hdr = EFI_HDR_WIN(*struct.unpack_from(header_fmt, buffer[off: off + header_size]))

        next_var_offset = off + efi_var_hdr.Size
        efi_var_buf = buffer[off: next_var_offset]
        efi_var_data = buffer[off + efi_var_hdr.DataOffset: off + efi_var_hdr.DataOffset + efi_var_hdr.DataSize]

        str_fmt = f'{efi_var_hdr.DataOffset - header_size:d}s'
        s, = struct.unpack(str_fmt, buffer[off + header_size: off + efi_var_hdr.DataOffset])
        efi_var_name = str(s, "utf-16-le", errors="replace").split(u'\u0000')[0]

        if efi_var_name not in variables.keys():
            variables[efi_var_name] = []
        #                                off, buf,         hdr,         data,         guid,                           attrs
        variables[efi_var_name].append((off, efi_var_buf, efi_var_hdr, efi_var_data, EFI_GUID_STR(efi_var_hdr.guid), efi_var_hdr.Attributes))

        if 0 == efi_var_hdr.Size:
            break
        off = next_var_offset

    return variables


def _handle_winerror(fn: str, msg: str, hr: int) -> None:
    _handle_error(f'{fn} failed: {msg} ({hr:d})', hr)


def _handle_error(err: str, hr: int = 0) -> None:
    if logger().DEBUG:
        logger().log_error(err)
    raise OsHelperError(err, hr)


class WindowsHelper(Helper):

    def __init__(self):
        super(WindowsHelper, self).__init__()

        self.os_system = platform.system()
        self.os_release = platform.release()
        self.os_version = platform.version()
        self.os_machine = platform.machine()
        self.os_uname = platform.uname()
        self.name = "WindowsHelper"
        win_ver = ""
        if "windows" == self.os_system.lower():
            win_ver = f"windows_{self.os_machine.lower()}"
            if ("5" == self.os_release):
                win_ver = "winxp"
            logger().log_debug(f'[helper] OS: {self.os_system} {self.os_release} {self.os_version}')

        self.use_existing_service = False

        self.win_ver = win_ver
        self.driver_handle = None
        self.device_file = str(DEVICE_FILE)

        c_int_p = POINTER(c_int)
        c_uint32_p = POINTER(c_uint32)

        # enable required SeSystemEnvironmentPrivilege privilege
        privilege = win32security.LookupPrivilegeValue(None, 'SeSystemEnvironmentPrivilege')
        token = win32security.OpenProcessToken(win32process.GetCurrentProcess(), win32security.TOKEN_READ | win32security.TOKEN_ADJUST_PRIVILEGES)
        win32security.AdjustTokenPrivileges(token, False, [(privilege, win32security.SE_PRIVILEGE_ENABLED)])
        win32api.CloseHandle(token)
        # import firmware variable API
        try:
            self.GetFirmwareEnvironmentVariable = kernel32.GetFirmwareEnvironmentVariableW
            self.GetFirmwareEnvironmentVariable.restype = c_int
            self.GetFirmwareEnvironmentVariable.argtypes = [c_wchar_p, c_wchar_p, c_void_p, c_int]
            self.SetFirmwareEnvironmentVariable = kernel32.SetFirmwareEnvironmentVariableW
            self.SetFirmwareEnvironmentVariable.restype = c_int
            self.SetFirmwareEnvironmentVariable.argtypes = [c_wchar_p, c_wchar_p, c_void_p, c_int]
        except AttributeError as msg:
            logger().log_warning("G[S]etFirmwareEnvironmentVariableW function doesn't seem to exist")

        try:
            self.NtEnumerateSystemEnvironmentValuesEx = windll.ntdll.NtEnumerateSystemEnvironmentValuesEx
            self.NtEnumerateSystemEnvironmentValuesEx.restype = c_int
            self.NtEnumerateSystemEnvironmentValuesEx.argtypes = [c_int, c_void_p, c_void_p]
        except AttributeError as msg:
            logger().log_warning("NtEnumerateSystemEnvironmentValuesEx function doesn't seem to exist")

        try:
            self.GetFirmwareEnvironmentVariableEx = kernel32.GetFirmwareEnvironmentVariableExW
            self.GetFirmwareEnvironmentVariableEx.restype = c_int
            self.GetFirmwareEnvironmentVariableEx.argtypes = [c_wchar_p, c_wchar_p, c_void_p, c_int, c_int_p]
            self.SetFirmwareEnvironmentVariableEx = kernel32.SetFirmwareEnvironmentVariableExW
            self.SetFirmwareEnvironmentVariableEx.restype = c_int
            self.SetFirmwareEnvironmentVariableEx.argtypes = [c_wchar_p, c_wchar_p, c_void_p, c_int, c_int]
        except AttributeError as msg:
            if logger().DEBUG:
                logger().log_warning("G[S]etFirmwareEnvironmentVariableExW function doesn't seem to exist")

        try:
            self.GetSystemFirmwareTbl = kernel32.GetSystemFirmwareTable
            self.GetSystemFirmwareTbl.restype = c_int
            self.GetSystemFirmwareTbl.argtypes = [c_int, c_int, c_void_p, c_int]
        except AttributeError as msg:
            logger().log_warning("GetSystemFirmwareTable function doesn't seem to exist")

        try:
            self.EnumSystemFirmwareTbls = kernel32.EnumSystemFirmwareTables
            self.EnumSystemFirmwareTbls.restype = c_int
            self.EnumSystemFirmwareTbls.argtypes = [c_int, c_void_p, c_int]
        except AttributeError as msg:
            logger().log_warning("GetSystemFirmwareTable function doesn't seem to exist")

        try:
            self.NtQuerySystemInformation = windll.ntdll.NtQuerySystemInformation
            self.NtQuerySystemInformation.restype = c_int
            self.NtQuerySystemInformation.argtypes = [c_uint32, c_void_p, c_uint32, c_uint32_p]
        except AttributeError as msg:
            logger().log_warning("NtQuerySystemInformation function doesn't seem to exist")

    def __del__(self):
        if self.driver_handle:
            win32api.CloseHandle(self.driver_handle)
            self.driver_handle = None

###############################################################################################
# Driver/service management functions
###############################################################################################

    def show_warning(self) -> None:
        logger().log("")
        logger().log_warning("*******************************************************************")
        logger().log_warning("Chipsec should only be used on test systems!")
        logger().log_warning("It should not be installed/deployed on production end-user systems.")
        logger().log_warning("See WARNING.txt")
        logger().log_warning("*******************************************************************")
        logger().log("")

    def create(self) -> bool:
        # check DRIVER_FILE_PATHS for the DRIVER_FILE_NAME
        self.driver_path = None
        for path in DRIVER_FILE_PATHS:
            driver_path = os.path.join(path, DRIVER_FILE_NAME)
            if os.path.isfile(driver_path):
                self.driver_path = driver_path
                logger().log_debug(f'[helper] Found driver in {driver_path}')
        if self.driver_path is None:
            logger().log_debug("[helper] CHIPSEC Windows Driver Not Found")
            raise Exception("CHIPSEC Windows Driver Not Found")

        self.show_warning()

        try:
            hscm = win32service.OpenSCManager(None, None, win32service.SC_MANAGER_ALL_ACCESS)  # SC_MANAGER_CREATE_SERVICE
        except win32service.error as err:
            _handle_winerror(err.args[1], err.args[2], err.args[0])
        logger().log_debug(f'[helper] service control manager opened (handle = {hscm})')
        logger().log_debug(f"[helper] driver path: '{os.path.abspath(self.driver_path)}'")

        try:
            hs = win32service.CreateService(
                hscm,
                SERVICE_NAME,
                DISPLAY_NAME,
                (win32service.SERVICE_QUERY_STATUS | win32service.SERVICE_START | win32service.SERVICE_STOP),
                win32service.SERVICE_KERNEL_DRIVER,
                win32service.SERVICE_DEMAND_START,
                win32service.SERVICE_ERROR_NORMAL,
                os.path.abspath(self.driver_path),
                None, 0, u"", None, None)
            if hs:
                logger().log_debug(f"[helper] Service '{SERVICE_NAME}' created (handle = 0x{int(hs):08X})")
        except win32service.error as err:
            if (winerror.ERROR_SERVICE_EXISTS == err.args[0]):
                logger().log_debug(f"[helper] Service '{SERVICE_NAME}' already exists: {err.args[2]} ({err.args[0]:d})")
                try:
                    hs = win32service.OpenService(hscm, SERVICE_NAME, (win32service.SERVICE_QUERY_STATUS | win32service.SERVICE_START | win32service.SERVICE_STOP))  # SERVICE_ALL_ACCESS
                except win32service.error as _err:
                    _handle_winerror(_err.args[1], _err.args[2], _err.args[0])
            else:
                _handle_winerror(err.args[1], err.args[2], err.args[0])

        finally:
            win32service.CloseServiceHandle(hs)
            win32service.CloseServiceHandle(hscm)

        return True

    def start(self) -> bool:

        self.use_existing_service = (win32serviceutil.QueryServiceStatus(SERVICE_NAME)[1] == win32service.SERVICE_RUNNING)

        if self.use_existing_service:
            self.driver_loaded = True
            logger().log_debug(f"[helper] Service '{SERVICE_NAME}' already running")
            logger().log_debug(f"[helper] Trying to connect to existing '{SERVICE_NAME}' service...")
        else:
            try:
                win32serviceutil.StartService(SERVICE_NAME)
                win32serviceutil.WaitForServiceStatus(SERVICE_NAME, win32service.SERVICE_RUNNING, 1)
                self.driver_loaded = True
                logger().log_debug(f"[helper] service '{SERVICE_NAME}' started")
            except pywintypes.error as err:
                _handle_error(f"Service '{SERVICE_NAME}' didn't start: {err.args[2]} ({err.args[0]:d})", err.args[0])
        self.driverpath = win32serviceutil.LocateSpecificServiceExe(SERVICE_NAME)
        self.driverpath = f'({self.driverpath})'
        return True

    def stop(self) -> bool:
        if self.use_existing_service:
            return True

        logger().log_debug(f"[helper] Stopping service '{SERVICE_NAME}'...")
        try:
            win32api.CloseHandle(self.driver_handle)
            self.driver_handle = None
            win32serviceutil.StopService(SERVICE_NAME)
        except pywintypes.error as err:
            if logger().DEBUG:
                logger().log_error(f'StopService failed: {err.args[2]} ({err.args[0]:d})')
            return False
        finally:
            self.driver_loaded = False

        try:
            win32serviceutil.WaitForServiceStatus(SERVICE_NAME, win32service.SERVICE_STOPPED, 1)
            logger().log_debug(f"[helper] Service '{SERVICE_NAME}' stopped")
        except pywintypes.error as err:
            if logger().DEBUG:
                logger().log_warning(f"Service '{SERVICE_NAME}' didn't stop: {err.args[2]} ({err.args[0]:d})")
            return False

        return True

    def delete(self) -> bool:
        if self.use_existing_service:
            return True

        if win32serviceutil.QueryServiceStatus(SERVICE_NAME)[1] != win32service.SERVICE_STOPPED:
            logger().log_warning(f"Cannot delete service '{SERVICE_NAME}' (not stopped)")
            return False

        logger().log_debug(f"[helper] Deleting service '{SERVICE_NAME}'...")
        try:
            win32serviceutil.RemoveService(SERVICE_NAME)
            logger().log_debug(f"[helper] Service '{SERVICE_NAME}' deleted")
        except win32service.error as err:
            if logger().DEBUG:
                logger().log_warning(f"RemoveService failed: {err.args[2]} ({err.args[0]:d})")
            return False

        return True

    def _get_driver_handle(self) -> 'PyHANDLE':
        # This is bad but DeviceIoControl fails occasionally if new device handle is not opened every time ;(
        if (self.driver_handle is not None) and (INVALID_HANDLE_VALUE != self.driver_handle):
            return self.driver_handle
        self.driver_handle = win32file.CreateFile(self.device_file, FILE_SHARE_READ | FILE_SHARE_WRITE, 0, None, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED, None)
        if (self.driver_handle is None) or (INVALID_HANDLE_VALUE == self.driver_handle):
            _handle_error(drv_hndl_error_msg, errno.ENXIO)
        else:
            logger().log_debug(f"[helper] Opened device '{DEVICE_FILE:.64}' (handle: {int(self.driver_handle):08X})")
        return self.driver_handle

    def check_driver_handle(self) -> bool:
        if (0x6 == kernel32.GetLastError()):
            win32api.CloseHandle(self.driver_handle)
            self.driver_handle = None
            self._get_driver_handle()
            logger().log_warning(f"Invalid handle: re-opened device '{self.device_file:.64}' (new handle: {int(self.driver_handle):08X})")
            return False
        return True

    #
    # Auxiliary functions
    #
    def get_threads_count(self) -> int:
        sum = 0
        proc_group_count = (kernel32.GetActiveProcessorGroupCount() & 0xFFFF)
        for grp in range(proc_group_count):
            procs = kernel32.GetActiveProcessorCount(grp)
            sum = sum + procs
        return sum

    #
    # Generic IOCTL call function
    #
    def _ioctl(self, ioctl_code: int, in_buf: bytes, out_length: int) -> Array:

        if not self.driver_loaded:
            _handle_error("chipsec kernel driver is not loaded (in native API mode?)")

        out_buf = (c_char * out_length)()
        self._get_driver_handle()
        try:
            out_buf = win32file.DeviceIoControl(self.driver_handle, ioctl_code, in_buf, out_length, None)
        except pywintypes.error as _err:
            err_status = _err.args[0] + 0x100000000
            if STATUS_PRIVILEGED_INSTRUCTION == err_status:
                err_msg = f'HW Access Violation: DeviceIoControl returned STATUS_PRIVILEGED_INSTRUCTION (0x{err_status:X})'
                if logger().DEBUG:
                    logger().log_error(err_msg)
                raise HWAccessViolationError(err_msg, err_status)
            else:
                _handle_error(f'HW Access Error: DeviceIoControl returned status 0x{err_status:X} ({_err.args[2]})', err_status)

        return out_buf

###############################################################################################
# Actual driver IOCTL functions to access HW resources
###############################################################################################

    def read_phys_mem(self, phys_address: int, length: int) -> bytes:
        out_length = length
        hi = (phys_address >> 32) & 0xFFFFFFFF
        lo = phys_address & 0xFFFFFFFF
        in_buf = struct.pack('3I', hi, lo, length)
        out_buf = self._ioctl(IOCTL_READ_PHYSMEM, in_buf, out_length)
        return bytes(out_buf)

    def write_phys_mem(self, phys_address: int, length: int, buf: AnyStr) -> int:
        hi = (phys_address >> 32) & 0xFFFFFFFF
        lo = phys_address & 0xFFFFFFFF
        in_buf = struct.pack('3I', hi, lo, length) + stringtobytes(buf)
        out_buf = self._ioctl(IOCTL_WRITE_PHYSMEM, in_buf, 4)
        return int.from_bytes(out_buf, 'little')

    # @TODO: Temporarily the same as read_phys_mem for compatibility
    def read_mmio_reg(self, phys_address: int, size: int) -> int:
        out_size = size
        logger().log_debug(f'[helper] -> read_mmio_reg( phys_address=0x{phys_address:X}, size={size} )')
        in_buf = struct.pack('3I', (phys_address >> 32) & 0xFFFFFFFF, phys_address & 0xFFFFFFFF, size)
        out_buf = self._ioctl(IOCTL_READ_MMIO, in_buf, out_size)
        if size == 8:
            value = struct.unpack('=Q', out_buf)[0]
        elif size == 4:
            value = struct.unpack('=I', out_buf)[0]
        elif size == 2:
            value = struct.unpack('=H', out_buf)[0]
        elif size == 1:
            value = struct.unpack('=B', out_buf)[0]
        else:
            value = 0
        return value

    def write_mmio_reg(self, phys_address: int, size: int, value: int) -> int:
        if size == 8:
            buf = struct.pack('=Q', value)
        elif size == 4:
            buf = struct.pack('=I', value & 0xFFFFFFFF)
        elif size == 2:
            buf = struct.pack('=H', value & 0xFFFF)
        elif size == 1:
            buf = struct.pack('=B', value & 0xFF)
        else:
            return False
        in_buf = struct.pack('3I', ((phys_address >> 32) & 0xFFFFFFFF), (phys_address & 0xFFFFFFFF), size) + buf
        out_buf = self._ioctl(IOCTL_WRITE_MMIO, in_buf, 4)
        return int.from_bytes(out_buf, 'little')

    def alloc_phys_mem(self, length: int, max_pa: int) -> Tuple[int, int]:
        in_length = 12
        out_length = 16
        in_buf = struct.pack('QI', max_pa, length)
        out_buf = self._ioctl(IOCTL_ALLOC_PHYSMEM, in_buf, out_length)
        (va, pa) = struct.unpack('2Q', out_buf)
        return (va, pa)

    def va2pa(self, va: int) -> Tuple[int, int]:
        error_code = 0
        in_length = 8
        out_length = 8
        in_buf = struct.pack('Q', va)
        out_buf = self._ioctl(IOCTL_GET_PHYSADDR, in_buf, out_length)
        pa = struct.unpack('Q', out_buf)[0]
        return (pa, error_code)

    #
    # HYPERCALL
    #
    def hypercall(self, rcx, rdx, r8, r9, r10, r11, rax, rbx, rdi, rsi, xmm_buffer):
        if self.os_machine == 'AMD64':
            arg_type = 'Q'
            out_length = 8
        else:
            arg_type = 'I'
            out_length = 4
        in_buf = struct.pack(f'<11{arg_type}', rcx, rdx, r8, r9, r10, r11, rax, rbx, rdi, rsi, xmm_buffer)
        out_buf = self._ioctl(IOCTL_HYPERCALL, in_buf, out_length)
        return struct.unpack(f'<{arg_type}', out_buf)[0]

    #
    # MAP_IO_SPACE
    #
    def map_io_space(self, physical_address, length, cache_type):
        out_length = 8
        in_buf = struct.pack('<3Q', physical_address, length, cache_type)
        out_buf = self._ioctl(IOCTL_MAP_IO_SPACE, in_buf, out_length)
        virtual_address = struct.unpack('<Q', out_buf)[0]
        return virtual_address

    #
    # FREE_PHYS_MEM
    #
    def free_phys_mem(self, physical_address):
        out_length = 8
        in_buf = struct.pack('<Q', physical_address)
        out_buf = self._ioctl(IOCTL_FREE_PHYSMEM, in_buf, out_length)
        return

    def read_msr(self, cpu_thread_id: int, msr_addr: int) -> Tuple[int, int]:
        out_length = 8
        in_buf = struct.pack('=2I', cpu_thread_id, msr_addr)
        out_buf = self._ioctl(IOCTL_RDMSR, in_buf, out_length)
        (eax, edx) = struct.unpack('2I', out_buf)
        return (eax, edx)

    def write_msr(self, cpu_thread_id: int, msr_addr: int, eax: int, edx: int) -> int:
        out_length = 0
        in_buf = struct.pack('=4I', cpu_thread_id, msr_addr, eax, edx)
        self._ioctl(IOCTL_WRMSR, in_buf, out_length)
        return True

    def read_pci_reg(self, bus: int, device: int, function: int, address: int, size: int) -> int:
        bdf = PCI_BDF(bus & 0xFFFF, device & 0xFFFF, function & 0xFFFF, address & 0xFFFF)
        out_length = size
        in_buf = struct.pack('4HB', bdf.BUS, bdf.DEV, bdf.FUNC, bdf.OFF, size)
        out_buf = self._ioctl(READ_PCI_CFG_REGISTER, in_buf, out_length)
        if 1 == size:
            value = struct.unpack('B', out_buf)[0]
        elif 2 == size:
            value = struct.unpack('H', out_buf)[0]
        else:
            value = struct.unpack('I', out_buf)[0]
        return value

    def write_pci_reg(self, bus: int, device: int, function: int, address: int, value: int, size: int) -> int:
        bdf = PCI_BDF(bus & 0xFFFF, device & 0xFFFF, function & 0xFFFF, address & 0xFFFF)
        out_length = 0
        in_buf = struct.pack('4HIB', bdf.BUS, bdf.DEV, bdf.FUNC, bdf.OFF, value, size)
        self._ioctl(WRITE_PCI_CFG_REGISTER, in_buf, out_length)
        return True

    def load_ucode_update(self, cpu_thread_id: int, ucode_update_buf: bytes) -> bool:
        out_length = 0
        in_buf = struct.pack('=IH', cpu_thread_id, len(ucode_update_buf)) + ucode_update_buf
        self._ioctl(IOCTL_LOAD_UCODE_PATCH, in_buf, out_length)
        return True

    def read_io_port(self, io_port: int, size: int) -> int:
        value = 0
        in_buf = struct.pack('=HB', io_port, size)
        out_buf = self._ioctl(IOCTL_READ_IO_PORT, in_buf, size)
        if 1 == size:
            value = struct.unpack('B', out_buf)[0]
        elif 2 == size:
            value = struct.unpack('H', out_buf)[0]
        else:
            value = struct.unpack('I', out_buf)[0]
        return value

    def write_io_port(self, io_port: int, value: int, size: int) -> bool:
        in_buf = struct.pack('=HIB', io_port, value, size)
        self._ioctl(IOCTL_WRITE_IO_PORT, in_buf, 0)
        return True

    def read_cr(self, cpu_thread_id: int, cr_number: int) -> int:
        value = 0
        in_buf = struct.pack('=HI', cr_number, cpu_thread_id)
        out_buf = self._ioctl(IOCTL_RDCR, in_buf, 8)
        value, = struct.unpack('=Q', out_buf)
        return value

    def write_cr(self, cpu_thread_id: int, cr_number: int, value: int) -> int:
        in_buf = struct.pack('=HQI', cr_number, value, cpu_thread_id)
        self._ioctl(IOCTL_WRCR, in_buf, 0)
        return True

    #
    # IDTR/GDTR/LDTR
    #
    def get_descriptor_table(self, cpu_thread_id: int, desc_table_code: int) -> Tuple[int, int, int]:
        in_buf = struct.pack('IB', cpu_thread_id, desc_table_code)
        out_buf = self._ioctl(IOCTL_GET_CPU_DESCRIPTOR_TABLE, in_buf, 18)
        (limit, base, pa) = struct.unpack('=HQQ', out_buf)
        return (limit, base, pa)

    #
    # EFI Variable API
    #
    def EFI_supported(self) -> bool:
        # kern32.GetFirmwareEnvironmentVariable with garbage parameters will cause GetLastError() == 1 reliably on a legacy system
        if self.GetFirmwareEnvironmentVariable is not None:
            self.GetFirmwareEnvironmentVariable("", "{00000000-0000-0000-0000-000000000000}", 0, 0)
            return win32api.GetLastError() != 1
        elif self.GetFirmwareEnvironmentVariableEx is not None:
            self.GetFirmwareEnvironmentVariableEx("", "{00000000-0000-0000-0000-000000000000}", 0, 0)
            return win32api.GetLastError() != 1
        else:
            return False

    def get_EFI_variable_full(self, name: str, guid: str, attrs: Optional[int] = None) -> Tuple[int, Optional[bytes], int]:
        status = 0  # EFI_SUCCESS
        length = 0
        efi_var = create_string_buffer(EFI_VAR_MAX_BUFFER_SIZE)
        if attrs is None:
            if self.GetFirmwareEnvironmentVariable is not None:
                logger().log_debug(f"[helper] -> GetFirmwareEnvironmentVariable( name='{name}', GUID='{{{guid}}}' )...")
                length = self.GetFirmwareEnvironmentVariable(name, f'{{{guid}}}', efi_var, EFI_VAR_MAX_BUFFER_SIZE)
        else:
            if self.GetFirmwareEnvironmentVariableEx is not None:
                pattrs = c_int(attrs)
                logger().log_debug(f"[helper] -> GetFirmwareEnvironmentVariableEx( name='{name}', GUID='{{{guid}}}', attrs = 0x{attrs:X} )...")
                length = self.GetFirmwareEnvironmentVariableEx(name, f'{{{guid}}}', efi_var, EFI_VAR_MAX_BUFFER_SIZE, pattrs)
        if (0 == length) or (efi_var is None):
            status = kernel32.GetLastError()
            if logger().DEBUG:
                logger().log_error(f'GetFirmwareEnvironmentVariable[Ex] returned error: {WinError()}')
            efi_var_data = None
            # raise WinError(errno.EIO,"Unable to get EFI variable")
        else:
            efi_var_data = bytes(efi_var[:length])

        return (status, efi_var_data, attrs)

    def get_EFI_variable(self, name: str, guid: str, attrs: Optional[int] = None) -> Optional[bytes]:
        (status, data, attributes) = self.get_EFI_variable_full(name, guid, attrs)
        return data

    def set_EFI_variable(self, name: str, guid: str, buffer: bytes, buffer_size: Optional[int], attrs: Optional[int]) -> int:
        var = bytes(0) if buffer is None else buffer
        var_len = len(var) if buffer_size is None else buffer_size
        ntsts = 0
        if isinstance(attrs, (str, bytes)):
            attrs_data = f'{bytestostring(attrs):\x00<8}'[:8]
            attrs = struct.unpack("Q", stringtobytes(attrs_data))[0]

        if attrs is None:
            if self.SetFirmwareEnvironmentVariable is not None:
                logger().log_debug(f"[helper] -> SetFirmwareEnvironmentVariable( name='{name}', GUID='{{{guid}}}', length=0x{var_len:X} )...")
                ntsts = self.SetFirmwareEnvironmentVariable(name, f'{{{guid}}}', var, var_len)
        else:
            if self.SetFirmwareEnvironmentVariableEx is not None:
                logger().log_debug(f"[helper] -> SetFirmwareEnvironmentVariableEx( name='{name}', GUID='{{{guid}}}', length=0x{var_len:X}, attrs=0x{attrs:X} )...")
                ntsts = self.SetFirmwareEnvironmentVariableEx(name, f'{{{guid}}}', var, var_len, attrs)
        if 0 != ntsts:
            status = 0  # EFI_SUCCESS
        else:
            status = kernel32.GetLastError()
            if logger().DEBUG:
                logger().log_error(f'SetFirmwareEnvironmentVariable[Ex] returned error: {WinError()}')
        return status

    def delete_EFI_variable(self, name: str, guid: str) -> int:
        return self.set_EFI_variable(name, guid, None, buffer_size=0, attrs=None)

    def list_EFI_variables(self, infcls: int = 2) -> Optional[Dict[str, List[Tuple[int, bytes, int, bytes, str, int]]]]:
        logger().log_debug(f'[helper] -> NtEnumerateSystemEnvironmentValuesEx( infcls={infcls:d} )...')
        efi_vars = create_string_buffer(EFI_VAR_MAX_BUFFER_SIZE)
        length = packl_ctypes(EFI_VAR_MAX_BUFFER_SIZE, 32)
        status = self.NtEnumerateSystemEnvironmentValuesEx(infcls, efi_vars, length)
        status = (((1 << 32) - 1) & status)
        if (0xC0000023 == status):
            retlength, = struct.unpack("<I", length)
            efi_vars = create_string_buffer(retlength)
            status = self.NtEnumerateSystemEnvironmentValuesEx(infcls, efi_vars, length)
        elif (0xC0000002 == status):
            if logger().DEBUG:
                logger().log_warning('NtEnumerateSystemEnvironmentValuesEx was not found (NTSTATUS = 0xC0000002)')
            logger().log_debug('[*] Your Windows does not expose UEFI Runtime Variable API. It was likely installed as legacy boot.\nTo use UEFI variable functions, chipsec needs to run in OS installed with UEFI boot (enable UEFI Boot in BIOS before installing OS)')
            return None
        if 0 != status:
            lasterror = kernel32.GetLastError()
            if (0xC0000001 == status):  # ERROR_NOACCESS: Invalid access to memory location.  https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-erref/18d8fbe8-a967-4f1c-ae50-99ca8e491d2d
                logger().log_warning('NtEnumerateSystemEnvironmentValuesEx was not successful')
                logger().log_debug(f'NTSTATUS = 0x{status:08X}, LastError = 0x{lasterror:X}')
                logger().log_warning('Looks like your version of Windows has restricted access to UEFI variables.\n\tTo use UEFI variable functions, chipsec needs to run in an older version of windows or in a different environment (Linux)')
            else:
                if logger().DEBUG:
                    logger().log_error(f'NtEnumerateSystemEnvironmentValuesEx failed (GetLastError = 0x{lasterror:X})')
                    logger().log_error(f'*** NTSTATUS: {status:08X}')
            return None
        logger().log_debug(f'[helper] len(efi_vars) = 0x{len(efi_vars):X} (should be 0x20000)')
        return getEFIvariables_NtEnumerateSystemEnvironmentValuesEx2(bytes(efi_vars))

    #
    # Interrupts
    #
    def send_sw_smi(self, cpu_thread_id: int, SMI_code_data: int, _rax: int, _rbx: int, _rcx: int, _rdx: int, _rsi: int, _rdi: int) -> Optional[Tuple[int, int, int, int, int, int, int]]:
        if (sys.maxsize < 2**32 and self.os_machine == 'AMD64') or (sys.maxsize > 2**32 and self.os_machine == 'i386'):
            logger().log(f"[helper] Python architecture must match OS architecture.  Run with {self.os_machine} architecture of python")
        out_length = struct.calcsize(_smi_msg_t_fmt)
        out_size = c_ulong(out_length)
        in_buf = struct.pack(_smi_msg_t_fmt, SMI_code_data, _rax, _rbx, _rcx, _rdx, _rsi, _rdi)
        out_buf = self._ioctl(IOCTL_SWSMI, in_buf, out_length)
        if out_buf:
            ret = struct.unpack(_smi_msg_t_fmt, out_buf)
        else:
            ret = None
        return ret

    def _get_handle_for_pid(self, pid: int = 0, ro: bool = True) -> int:
        if pid == 0:
            pHandle = win32process.GetCurrentProcess()
        else:
            flags = win32con.PROCESS_QUERY_INFORMATION
            if not ro:
                flags |= win32con.PROCESS_SET_INFORMATION
            try:
                pHandle = win32api.OpenProcess(flags, 0, pid)
            except pywintypes.error as e:
                logger().log("unable to open a process handle")
                raise ValueError(e)
        return pHandle

    def set_affinity(self, value: int) -> Optional[int]:
        pHandle = self._get_handle_for_pid(0, False)
        current = win32process.GetProcessAffinityMask(pHandle)[0]
        try:
            win32process.SetProcessAffinityMask(pHandle, current)
        except win32process.error as e:
            logger().log("unable to set process affinity")
            raise ValueError(e)
        return current

    def get_affinity(self) -> Optional[int]:
        pHandle = self._get_handle_for_pid()
        try:
            return win32process.GetProcessAffinityMask(pHandle)[0]
        except win32process.error as e:
            logger().log("unable to get the running cpu")
            raise ValueError(e)

    #
    # CPUID
    #
    def cpuid(self, eax: int, ecx: int) -> Tuple[int, int, int, int]:
        out_length = 16
        in_buf = struct.pack('=2I', eax, ecx)
        out_buf = self._ioctl(IOCTL_CPUID, in_buf, out_length)
        (eax, ebx, ecx, edx) = struct.unpack('4I', out_buf)
        return (eax, ebx, ecx, edx)

    def enum_ACPI_tables(self) -> Optional[Array]:
        table_size = 36
        tBuffer = create_string_buffer(table_size)
        retVal = self.EnumSystemFirmwareTbls(FirmwareTableProviderSignature_ACPI, tBuffer, table_size)
        if retVal == 0:
            if logger().DEBUG:
                logger().log_error(f'EnumSystemFirmwareTbls() returned error: {WinError()}')
            return None
        if retVal > table_size:
            table_size = retVal
            tBuffer = create_string_buffer(table_size)
            retVal = self.EnumSystemFirmwareTbls(FirmwareTableProviderSignature_ACPI, tBuffer, table_size)
        tables_array = [tBuffer[i:i+4] for i in range(0, retVal, 4)]
        return tables_array

    # ACPI access is implemented through ACPI HAL rather than through kernel module
    def get_ACPI_table(self, table_name: str) -> Optional[Array]:
        table_size = 36
        tBuffer = create_string_buffer(table_size)
        tbl = struct.unpack("<I", bytes(table_name, 'ascii'))[0]
        retVal = self.GetSystemFirmwareTbl(FirmwareTableProviderSignature_ACPI, tbl, tBuffer, table_size)
        if retVal == 0:
            if logger().DEBUG:
                logger().log_error(f'GetSystemFirmwareTable({table_name}) returned error: {WinError()}')
            return None
        if retVal > table_size:
            table_size = retVal
            tBuffer = create_string_buffer(table_size)
            retVal = self.GetSystemFirmwareTbl(FirmwareTableProviderSignature_ACPI, tbl, tBuffer, table_size)
        return tBuffer[:retVal]

    #
    # IOSF Message Bus access
    #

    def msgbus_send_read_message(self, mcr, mcrx):
        raise UnimplementedAPIError("msgbus_send_read_message")

    def msgbus_send_write_message(self, mcr, mcrx, mdr):
        raise UnimplementedAPIError("msgbus_send_write_message")

    def msgbus_send_message(self, mcr, mcrx, mdr):
        raise UnimplementedAPIError("msgbus_send_message")

    #
    # Speculation control
    #
    def retpoline_enabled(self) -> bool:
        # See https://docs.microsoft.com/en-us/windows/win32/api/winternl/nf-winternl-ntquerysysteminformation
        speculation_control = c_uint32(0)
        SystemSpeculationControlInformation = 0xC9
        SpecCtrlRetpolineEnabled = 0x4000
        self.NtQuerySystemInformation(SystemSpeculationControlInformation, addressof(speculation_control), sizeof(speculation_control), None)
        return bool(speculation_control.value & SpecCtrlRetpolineEnabled)

#
# Get instance of this OS helper
#


def get_helper() -> WindowsHelper:
    return WindowsHelper()



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2024, AMD Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# coudrant@amd.com
#


"""
CPU SMM Addr

This module checks to see that SMMMask has Tseg and Aseg programmed correctly.  It also verifies that CPU access to SMM
is blocked while not in SMM.

Usage:
    ``chipsec_main -m common.smm_addr``

Examples:
    >>> chipsec_main.py -m common.smm_addr

Registers used:
    AMD:
    - SMMMASK.TMTypeDram
    - SMMMASK.AMTypeDram
    - SMMMASK.TValid
    - SMMMASK.AValid
    - SMM_BASE


"""

from chipsec.module_common import BaseModule, BIOS, SMM
from chipsec.library.returncode import ModuleResult
from chipsec.hal.msr import MemType
from typing import List

TAGS = [BIOS, SMM]
METADATA_TAGS = ['OPENSOURCE', 'IA', 'COMMON', 'SMM_ADDR']


class smm_addr(BaseModule):

    def __init__(self):
        BaseModule.__init__(self)

    def is_supported(self) -> bool:
        if self.cs.is_amd():
            smmaddr_exist = self.cs.register.is_defined('SMMADDR')
            smmmask_exist = self.cs.register.is_defined('SMMMASK')
            if smmaddr_exist and smmmask_exist:
                return True
        self.logger.log_information('Required registers are not defined for this platform.  Skipping module.')
        return False

    def check_SMMMask(self) -> int:

        if self.cs.cpu.check_SMRR_supported():
            self.logger.log_good("OK. SMMMask range protection is supported")
        else:
            self.logger.log_not_applicable("CPU does not support SMMMask range protection of SMRAM")
            self.result.setStatusBit(self.result.status.NOT_APPLICABLE)
            self.res = self.result.getReturnCode(ModuleResult.NOTAPPLICABLE)
        #
        # SMMMask are supported
        #
        smm_addr_ok = True

        #
        # 2. Check SMMMASK is programmed correctly (on CPU0)
        #
        self.logger.log('')
        self.logger.log("[*] Checking SMMMask range base programming..")

        cpu0_smmmask = self.cs.register.read('SMMMASK')
        self.cs.register.print('SMMMASK', cpu0_smmmask)
        tseg_type = self.cs.register.get_field('SMMMASK', cpu0_smmmask, 'TMTYPEDRAM')
        aseg_type = self.cs.register.get_field('SMMMASK', cpu0_smmmask, 'AMTYPEDRAM')

        if tseg_type in MemType:
            self.logger.log(f"[*] TSEG range memory type is {MemType[tseg_type]}")
        else:
            smm_addr_ok = False
            self.logger.log_bad(f"TSEG range memory type 0x{tseg_type:X} is invalid")

        if aseg_type in MemType:
            self.logger.log(f"[*] ASEG range memory type is {MemType[aseg_type]}")
        else:
            self.logger.log_bad(f"ASEG range memory type 0x{aseg_type:X} is invalid")

        if smm_addr_ok:
            self.logger.log_good("OK so far. ASEG and TESEG are programmed correctly")

        #
        # 3. Check TSeg and ASeg are Valid on CPU0
        #
        self.logger.log('')
        self.logger.log("[*] Checking SMM Mask Aseg and Tseg validity..")
        smmmask = self.cs.register.read('SMMMASK')
        self.cs.register.print('SMMMASK', smmmask)
        tsegvalid = self.cs.register.get_field('SMMMASK', smmmask, 'TVALID')
        asegvalid = self.cs.register.get_field('SMMMASK', smmmask, 'AVALID')

        if (not bool(tsegvalid) or not bool(asegvalid)):
            smm_addr_ok = False
            self.logger.log_bad("SMMMask range is not enabled")

        if smm_addr_ok:
            self.logger.log_good("OK so far. SMMMask range is enabled")

        #
        # 4. Verify that SMMMask_BASE/MASK MSRs have the same values on all logical CPUs
        #
        self.logger.log('')
        self.logger.log("[*] Verifying that SMM range base & mask are the same on all logical CPUs..")
        smmbase  = []
        for tid in range(self.cs.msr.get_cpu_thread_count()):
            smmbase.append(self.cs.register.read('SMM_BASE', tid))
            smmmask = self.cs.register.read('SMMMASK', tid)
            self.logger.log(f"[CPU{tid:d}] SMMMask = {smmbase[tid]:016X}, SMMMask = {smmmask:016X}")

            if (smmmask != cpu0_smmmask):
                smm_addr_ok = False
                self.logger.log_bad("SMMMask range base/mask do not match on all logical CPUs")
                break

        if smm_addr_ok:
            self.logger.log_good("OK so far. SMMMask range base/mask match on all logical CPUs")

        #
        # 5. Reading from & writing to SMMMask Dram physical address
        # writes should be dropped, reads should return all F's
        #

        for tid_base in smmbase:
            self.logger.log(f"[*] Trying to read memory at SMM base 0x{tid_base:08X}..")
            ok = 0xFFFFFFFF == self.cs.mem.read_physical_mem_dword(tid_base)
            if not ok:
                self.logger.log_bad("Able to read SMM base at 0x{tid_base:08X}..")
                break

        smm_addr_ok = smm_addr_ok and ok
        if ok:
            self.logger.log_passed("SMMMask reads are blocked in non-SMM mode")  # return all F's
        else:
            self.logger.log_failed("SMMMask reads are not blocked in non-SMM mode")  # all F's are not returned

        self.logger.log('')
        if not smm_addr_ok:
            res = ModuleResult.FAILED
            self.result.setStatusBit(self.result.status.CONFIGURATION)
            self.logger.log_failed("SMMMask protection against cache attack is not configured properly")
        else:
            res = ModuleResult.PASSED
            self.logger.log_passed("SMMask protection against cache attack is properly configured")

        res = ModuleResult.PASSED

        return self.result.getReturnCode(res)

    # --------------------------------------------------------------------------
    # run( module_argv )
    # Required function: run here all tests from this module
    # --------------------------------------------------------------------------
    def run(self, module_argv: List[str]) -> int:
        self.logger.start_test("CPU SMM Cache Poisoning / System Management Mode Mask Registers")
        self.res = self.check_SMMMask()
        return self.res



#!/usr/bin/env python3
# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Standalone utility
"""

import argparse
import importlib
import os
import sys
from time import time
from typing import Optional, Dict, Any, Sequence

from chipsec.helper.oshelper import helper
from chipsec.library.logger import logger, level
from chipsec.library.banner import print_banner, print_banner_properties
from chipsec.library.exceptions import UnknownChipsetError
from chipsec.library.options import Options
from chipsec.testcase import ExitCode
from chipsec.chipset import cs
from chipsec.library.file import get_main_dir
from chipsec.library.defines import get_version, get_message, os_version

CMD_OPTS_WIDTH = {'byte': 0x1, 'word': 0x2, 'dword': 0x4}


def is_option_valid_width(width_op):
    return (width_op.lower() in CMD_OPTS_WIDTH.keys())


def get_option_width(width_op):
    width_op = width_op.lower()
    return CMD_OPTS_WIDTH.get(width_op, 0)


def import_cmds() -> Dict[str, Any]:
    """Determine available chipsec_util commands"""
    cmds_dir = os.path.join(get_main_dir(), "chipsec", "utilcmd")
    cmds = [i[:-3] for i in os.listdir(cmds_dir) if i[-3:] == ".py" and not i[:2] == "__"]

    if logger().DEBUG:
        logger().log('[CHIPSEC] Loaded command-line extensions:')
        logger().log(f'   {cmds}')
    module = None
    commands = {}
    for cmd in cmds:
        try:
            cmd_path = f'chipsec.utilcmd.{cmd}'
            module = importlib.import_module(cmd_path)
            cu = getattr(module, 'commands')
            commands.update(cu)
        except ImportError as msg:
            # Display the import error and continue to import commands
            logger().log_error(f"Exception occurred during import of {cmd}: '{str(msg)}'")
            continue
    commands.update({"help": ""})
    return commands


def parse_args(argv: Sequence[str]) -> Optional[Dict[str, Any]]:
    """Parse the arguments provided on the command line."""
    options = Options()

    default_helper = options.get_section_data('Util_Config', 'default_helper', None)
    global_usage = "Additional arguments for specific command.\n\n All numeric values are in hex\n<width> is in {1, byte, 2, word, 4, dword}\n\n"
    cmds = import_cmds()
    parser = argparse.ArgumentParser(usage='%(prog)s [options] <command>', add_help=False)
    options = parser.add_argument_group('Options')
    options.add_argument('-h', '--help', dest='show_help', help="Show this message and exit", action='store_true')
    options.add_argument('-v', '--verbose', help='Verbose logging', action='store_true')
    options.add_argument('--hal', help='HAL logging', action='store_true')
    options.add_argument('-d', '--debug', help='Debug logging', action='store_true')
    options.add_argument('-vv', '--vverbose', help='Very verbose logging (Verbose + HAL + Debug)', action='store_true')
    options.add_argument('-l', '--log', help='Output to log file')
    options.add_argument('-p', '--platform', dest='_platform', help='Explicitly specify platform code', choices=cs().Cfg.proc_codes, type=str.upper)
    options.add_argument('--pch', dest='_pch', help='Explicitly specify PCH code', choices=cs().Cfg.pch_codes, type=str.upper)
    options.add_argument('-n', '--no_driver', dest='_no_driver', action='store_true',
                         help="Chipsec won't need kernel mode functions so don't load chipsec driver")
    options.add_argument('-i', '--ignore_platform', dest='_ignore_platform', action='store_true',
                         help='Run chipsec even if the platform is not recognized (Deprecated)')
    options.add_argument('--helper', dest='_helper', help='Specify OS Helper', choices=helper().get_available_helpers(), default=default_helper)
    options.add_argument('-nb', '--no_banner', dest='_show_banner', action='store_false', help="Chipsec won't display banner information")
    options.add_argument('--skip_config', dest='_load_config', action='store_false', help='Skip configuration and driver loading')
    options.add_argument('-nl', dest='_autolog_disable', action='store_true', help="Chipsec won't save logs automatically")
    options.add_argument('-rc', dest='_return_codes', help='Return codes mode', action='store_true')
    options.add_argument('_cmd', metavar='Command', nargs='?', choices=sorted(cmds.keys()), type=str.lower, default="help",
                         help=f"Util command to run: {{{','.join(sorted(cmds.keys()))}}}")
    options.add_argument('_cmd_args', metavar='Command Args', nargs=argparse.REMAINDER, help=global_usage)
    par = vars(parser.parse_args(argv))

    if par['_cmd'] == 'help' or par['show_help']:
        if par['_show_banner']:
            print_banner(argv, get_version(), get_message())
        parser.print_help()
        return None
    else:
        par['commands'] = cmds
        return par


class ChipsecUtil:

    def __init__(self, switches, argv):
        self.logger = logger()
        self.logger.UTIL_TRACE = True
        self.commands = switches['commands']
        self.__dict__.update(switches)
        self.argv = argv
        self.parse_switches()
        self._cs = cs()

    def parse_switches(self) -> None:
        self.logger.set_log_level(self.verbose, self.hal, self.debug, self.vverbose)
        if self.log:
            self.logger.set_log_file(self.log)
            self._autolog_disable = True
        if self._autolog_disable is False:
            self.logger.set_autolog_file("util")
        if self._return_codes:
            self.logger.log_warning("Return codes feature is currently Work in Progress!!!")
            self._cs.using_return_codes = True

        if not self._cmd_args:
            self._cmd_args = ["--help"]

    ##################################################################################
    # Entry point
    ##################################################################################

    def main(self) -> int:
        """Receives and executes the commands"""
        if self._show_banner:
            print_banner(self.argv, get_version(), get_message())

        comm = self.commands[self._cmd](self._cmd_args, cs=self._cs)
        comm.parse_arguments()
        reqs = comm.requirements()
        if reqs.load_driver() and self._no_driver:
            self.logger.log("Cannot run command without a driver loaded.", level.ERROR)
            return ExitCode.ERROR

        if reqs.load_config() and not self._load_config:
            self.logger.log("Cannot run command without a config loaded. Please run with -p and/or --pch if needed.", level.ERROR)
            return ExitCode.ERROR

        try:
            self._cs.init(self._platform, self._pch, self._helper, reqs.load_driver(), reqs.load_config(), self._ignore_platform)
        except UnknownChipsetError as msg:
            self.logger.log_error(f'Platform is not supported ({str(msg)}).')
            self.logger.log_error('To specify a cpu please use -p command-line option')
            self.logger.log_error('To specify a pch please use --pch command-line option\n')
            self.logger.log_error('If the correct configuration is not loaded, results should not be trusted.')
            return ExitCode.EXCEPTION
        except Exception as msg:
            self.logger.log(str(msg), level.ERROR)
            return ExitCode.EXCEPTION

        if self._show_banner:
            print_banner_properties(self._cs, os_version())

        self.logger.log(f"[CHIPSEC] Executing command '{self._cmd}' with args {self._cmd_args}\n")

        try:
            comm.set_up()
        except Exception as msg:
            self.logger.log_error(msg)
            return

        t = time()
        if comm.prerun():
            comm.run()
        self.logger.log(f"[CHIPSEC] Time elapsed {time()-t:.3f}")

        comm.tear_down()
        if reqs.load_driver() and not self._no_driver:
            self._cs.destroy_helper()
        return comm.ExitCode


def run(cli_cmd: str = '') -> int:
    cli_cmds = []
    if cli_cmd:
        cli_cmds = cli_cmd.strip().split(' ')
    return main(cli_cmds)


def main(argv: Sequence[str] = sys.argv[1:]) -> int:
    par = parse_args(argv)
    if par is not None:
        chipsecMain = ChipsecUtil(par, argv)
        return chipsecMain.main()
    return ExitCode.OK


if __name__ == "__main__":
    sys.exit(main())



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2023, Intel Corporation

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from inspect import stack
from json import dumps, loads
import os
from typing import Any, Dict, List, Optional, Tuple, TYPE_CHECKING
if TYPE_CHECKING:
    from ctypes import Array
from datetime import datetime

from chipsec.library.logger import logger
from chipsec.library.defines import bytestostring
from chipsec.helper.basehelper import Helper
from chipsec.helper.oshelper import helper
from chipsec.library.file import read_file, write_file


class RecordHelper(Helper):
    def __init__(self, filename: str = "", subhelper: str = ""):
        super(RecordHelper, self).__init__()
        self.os_system = "test_helper"
        self.os_release = "0"
        self.os_version = "0"
        self.os_machine = "test"
        self.driver_created = False
        self.driver_loaded = False
        self.name = "FileHelper"

        dtstr = datetime.now().strftime("%Y%m%d%H%M%S")
        self.default_file_name = f"recording{dtstr}.json"
        self.default_file_location = os.path.join("chipsec", "helper", "record")
        self._subhelper = helper().get_helper(subhelper) if subhelper else helper().get_default_helper()
        logger().log(f"Using subhelper: {self._subhelper.name}")
        self._filename = filename if filename else os.path.join(self.default_file_location, self.default_file_name)
        self._data = {}

    def switch_subhelper(self, newhelper: Helper) -> None:
        if self.driver_loaded:
            self._subhelper.stop()
        if self.driver_created:
            self._subhelper.delete()
        self._subhelper = newhelper
        logger().log(f"Switched subhelper to: {self._subhelper.name}")

    def _add_element(self, cmd: str, args: Tuple, ret: Any) -> None:
        try:
            margs = f"({','.join(str(i) for i in args)})"
        except Exception:
            margs = str(args)
        if isinstance(ret, bytes):
            ret = bytestostring(ret)
        if str(cmd) in self._data:
            if margs in self._data[str(cmd)]:
                # using insert opposed to append so that it creates last in first out when using pop command within getElement
                self._data[str(cmd)][margs].insert(0, str(ret))
            else:
                self._data[str(cmd)][margs] = [str(ret)]
        else:
            self._data[str(cmd)] = {margs: [str(ret)]}

    def _save(self) -> None:
        js = dumps(self._data, sort_keys=False, indent=2, separators=(',', ': '))
        write_file(self._filename, js)

    def _load(self) -> None:
        if os.path.isfile(self._filename):
            file_data = read_file(self._filename)
            if file_data == 0:
                self._data = {}
            try:
                self._data = loads(file_data)
            except Exception:
                self._data = {}

    def _call_subhelper(self, *myargs):
        fname= stack()[1][3]  # gets the name of the function that called _call_subhelper
        func = getattr(self._subhelper, fname)
        err = None
        try:
            ret = func(*myargs)
        except Exception as e:
            ret = f'!! {type(e)}: {e}'
            err = e
        self._add_element(fname, myargs, ret)
        if err is not None:
            raise err
        return ret

    def create(self) -> bool:
        self.driver_created = True
        return self._subhelper.create()

    def start(self) -> bool:
        self._load()  # Load file if it exists
        self.driver_loaded = True
        return self._subhelper.start()

    def stop(self) -> bool:
        self._save()
        self.driver_loaded = False
        return self._subhelper.stop()

    def delete(self) -> bool:
        self.driver_created = False
        return self._subhelper.delete()

    def read_pci_reg(self, bus: int, device: int, function: int, offset: int, size: int) -> int:
        return self._call_subhelper(bus, device, function, offset, size)

    def write_pci_reg(self, bus: int, device: int, function: int, offset: int, value: int, size: int) -> int:
        return self._call_subhelper(bus, device, function, offset, value, size)

    def read_mmio_reg(self, phys_address: int, size: int) -> int:
        return self._call_subhelper(phys_address, size)

    def write_mmio_reg(self, phys_address: int, size: int, value: int) -> int:
        return self._call_subhelper(phys_address, size, value)

    #
    # physical_address is 64 bit integer
    #
    def read_phys_mem(self, phys_address: int, length: int) -> bytes:
        return self._call_subhelper(phys_address, length)

    def write_phys_mem(self, phys_address: int, length: int, buf: bytes) -> int:
        return self._call_subhelper(phys_address, length, buf)

    def alloc_phys_mem(self, length: int, max_phys_address: int) -> Tuple[int, int]:
        return self._call_subhelper(length, max_phys_address)

    def free_phys_mem(self, physical_address: int) -> Optional[int]:
        return self._call_subhelper(physical_address)

    def va2pa(self, virtual_address: int) -> Tuple[int, int]:
        return self._call_subhelper(virtual_address)

    def map_io_space(self, physical_address: int, length: int, cache_type: int) -> int:
        return self._call_subhelper(physical_address, length, cache_type)

    #
    # Read/Write I/O port
    #
    def read_io_port(self, io_port: int, size: int) -> int:
        return self._call_subhelper(io_port, size)

    def write_io_port(self, io_port: int, value: int, size: int) -> int:
        return self._call_subhelper(io_port, value, size)

    #
    # Read/Write CR registers
    #
    def read_cr(self, cpu_thread_id: int, cr_number: int) -> int:
        return self._call_subhelper(cpu_thread_id, cr_number)

    def write_cr(self, cpu_thread_id: int, cr_number: int, value: int) -> int:
        return self._call_subhelper(cpu_thread_id, cr_number, value)

    #
    # Read/Write MSR on a specific CPU thread
    #
    def read_msr(self, cpu_thread_id: int, msr_addr: int) -> Tuple[int, int]:
        return self._call_subhelper(cpu_thread_id, msr_addr)

    def write_msr(self, cpu_thread_id: int, msr_addr: int, eax: int, edx: int) -> int:
        return self._call_subhelper(cpu_thread_id, msr_addr, eax, edx)

    #
    # Load CPU microcode update on a specific CPU thread
    #
    def load_ucode_update(self, cpu_thread_id: int, ucode_update_buf: bytes) -> bool:
        return self._call_subhelper(cpu_thread_id, ucode_update_buf)

    #
    # Read IDTR/GDTR/LDTR on a specific CPU thread
    #
    def get_descriptor_table(self, cpu_thread_id: int, desc_table_code: int) -> Optional[Tuple[int, int, int]]:
        return self._call_subhelper(cpu_thread_id, desc_table_code)

    #
    # EFI Variable API
    #
    def EFI_supported(self) -> bool:
        return self._call_subhelper()

    def get_EFI_variable(self, name: str, guid: str) -> Optional[bytes]:
        return self._call_subhelper(name, guid)

    def set_EFI_variable(self, name: str, guid: str, data: bytes, datasize: Optional[int], attrs: Optional[int]) -> Optional[int]:
        return self._call_subhelper(name, guid, data, datasize, attrs)

    def delete_EFI_variable(self, name: str, guid: str) -> Optional[int]:
        return self._call_subhelper(name, guid)

    def list_EFI_variables(self) -> Optional[Dict[str, List['EfiVariableType']]]:
        return self._call_subhelper()

    #
    # ACPI
    #

    def get_ACPI_table(self, table_name: str) -> Optional['Array']:
        return self._call_subhelper(table_name)

    def enum_ACPI_tables(self) -> Optional['Array']:
        return self._call_subhelper()

    #
    # CPUID
    #
    def cpuid(self, eax: int, ecx: int) -> Tuple[int, int, int, int]:
        return self._call_subhelper(eax, ecx)

    #
    # IOSF Message Bus access
    #
    def msgbus_send_read_message(self, mcr: int, mcrx: int) -> Optional[int]:
        return self._call_subhelper(mcr, mcrx)

    def msgbus_send_write_message(self, mcr: int, mcrx: int, mdr: int) -> None:
        return self._call_subhelper(mcr, mcrx, mdr)

    def msgbus_send_message(self, mcr: int, mcrx: int, mdr: Optional[int]) -> Optional[int]:
        return self._call_subhelper(mcr, mcrx, mdr)

    #
    # Affinity
    #
    def get_affinity(self) -> Optional[int]:
        return self._call_subhelper()

    def set_affinity(self, value: int) -> Optional[int]:
        return self._call_subhelper(value)

    #
    # Logical CPU count
    #
    def get_threads_count(self) -> int:
        return self._call_subhelper()

    #
    # Send SW SMI
    #
    def send_sw_smi(self, cpu_thread_id: int, SMI_code_data: int, _rax: int, _rbx: int, _rcx: int, _rdx: int, _rsi: int, _rdi: int) -> Optional[int]:
        return self._call_subhelper(cpu_thread_id, SMI_code_data, _rax, _rbx, _rcx, _rdx, _rsi, _rdi)

    #
    # Hypercall
    #
    def hypercall(self, rcx: int, rdx: int, r8: int, r9: int, r10: int, r11: int, rax: int, rbx: int, rdi: int, rsi: int, xmm_buffer: int) -> int:
        return self._call_subhelper(rcx, rdx, r8, r9, r10, r11, rax, rbx, rdi, rsi, xmm_buffer)

    #
    # Speculation control
    #
    def retpoline_enabled(self) -> bool:
        return self._call_subhelper()


def get_helper():
    return RecordHelper()




# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2024, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#
""""
To execute: python[3] -m unittest tests.utilcmd.vmem_cmd.test_vmem_cmd
"""

import unittest
import os

from chipsec.library.file import get_main_dir
from unittest.mock import patch
from tests.utilcmd.run_chipsec_util import setup_run_destroy_util_get_log_output
from chipsec.testcase import ExitCode


class TestVmemUtilCmd(unittest.TestCase):
    def test_readval(self) -> None:
        init_replay_file = os.path.join(get_main_dir(), "tests", "utilcmd", "adlenumerate.json")
        vmem_dump_replay_file = os.path.join(get_main_dir(), "tests", "utilcmd", "vmem_cmd", "vmem_cmd_readval_1.json")
        retval, log_call_string = setup_run_destroy_util_get_log_output(init_replay_file, "mem", "readval 0xFED40000 dword", util_replay_file=vmem_dump_replay_file)
        self.assertEqual(retval, ExitCode.OK)

    @patch('chipsec.utilcmd.vmem_cmd.write_file')
    def test_read(self, mock_write_file) -> None:
        init_replay_file = os.path.join(get_main_dir(), "tests", "utilcmd", "adlenumerate.json")
        vmem_dump_replay_file = os.path.join(get_main_dir(), "tests", "utilcmd", "vmem_cmd", "vmem_cmd_read_1.json")
        retval, log_call_string = setup_run_destroy_util_get_log_output(init_replay_file, "vmem", "read 0x41E 0x10 mock_buffer.bin", util_replay_file=vmem_dump_replay_file)
        self.assertRegex(log_call_string, r'41E.*0x10')
        self.assertEqual(retval, ExitCode.OK)

    def test_writeval(self) -> None:
        init_replay_file = os.path.join(get_main_dir(), "tests", "utilcmd", "adlenumerate.json")
        vmem_dump_replay_file = os.path.join(get_main_dir(), "tests", "utilcmd", "vmem_cmd", "vmem_cmd_writeval_1.json")
        retval, log_call_string = setup_run_destroy_util_get_log_output(init_replay_file, "vmem", "writeval 0xA0000 dword 0x9090CCCC", util_replay_file=vmem_dump_replay_file)
        self.assertEqual(retval, ExitCode.OK)

    def test_write(self) -> None:
        init_replay_file = os.path.join(get_main_dir(), "tests", "utilcmd", "adlenumerate.json")
        vmem_dump_replay_file = os.path.join(get_main_dir(), "tests", "utilcmd", "vmem_cmd", "vmem_cmd_write_1.json")
        retval, log_call_string = setup_run_destroy_util_get_log_output(init_replay_file, "vmem", "write 0x100000000 0x10 000102030405060708090A0B0C0D0E0F", util_replay_file=vmem_dump_replay_file)
        self.assertEqual(retval, ExitCode.OK)

    def test_allocate(self) -> None:
        init_replay_file = os.path.join(get_main_dir(), "tests", "utilcmd", "adlenumerate.json")
        vmem_dump_replay_file = os.path.join(get_main_dir(), "tests", "utilcmd", "vmem_cmd", "vmem_cmd_allocate_1.json")
        retval, log_call_string = setup_run_destroy_util_get_log_output(init_replay_file, "vmem", "allocate 0x1000", util_replay_file=vmem_dump_replay_file)
        self.assertEqual(retval, ExitCode.OK)

    def test_search(self) -> None:
        init_replay_file = os.path.join(get_main_dir(), "tests", "utilcmd", "adlenumerate.json")
        vmem_dump_replay_file = os.path.join(get_main_dir(), "tests", "utilcmd", "vmem_cmd", "vmem_cmd_allocate_1.json")
        retval, log_call_string = setup_run_destroy_util_get_log_output(init_replay_file, "vmem", "search 0xF0000 0x10 _SM_", util_replay_file=vmem_dump_replay_file)
        self.assertEqual(retval, ExitCode.OK)

    def test_getphys(self) -> None:
        init_replay_file = os.path.join(get_main_dir(), "tests", "utilcmd", "adlenumerate.json")
        vmem_dump_replay_file = os.path.join(get_main_dir(), "tests", "utilcmd", "vmem_cmd", "vmem_cmd_getphys_1.json")
        retval, log_call_string = setup_run_destroy_util_get_log_output(init_replay_file, "vmem", "getphys 0xFFFF9752F773F000", util_replay_file=vmem_dump_replay_file)
        self.assertEqual(retval, ExitCode.OK)
